#ifndef LIB_MQLPLUS_BASE_OBJ_LIST_TEMPLATES_MQH_INCLUDED
#define LIB_MQLPLUS_BASE_OBJ_LIST_TEMPLATES_MQH_INCLUDED
#property version "1.0";
/**********************************************************************************
 * Copyright (C) 2010-2022 Dominik Egert <info@freie-netze.de>
 *
 * This file is the objects include file.
 *
 * MQLplus, including this file may not be copied and/or distributed
 * without explecit permit by the author.
 * Author Dominik Egert / Freie Netze UG.
 **********************************************************************************
 *
 *  Version: 1.0
 *  State: production
 *
 *  File information
 *  ================
 *
 *
 *
*/
#ifdef DBG_MSG_TRACE_FILE_LOADER
    DBG_MSG_TRACE_FILE_LOADER;

#endif



// Incluse element handler
#include "lib_base_node_element.mqh"



/*********************************************************************************************************************************************************/
/*                                                                                                                                                       */
/* MQLplus data structures                                                                                                                               */
/*                                                                                                                                                       */
/*********************************************************************************************************************************************************/


    ///////////////////////////////////////
    //
    //  Linked list object
    //

        // Base object
        template <typename USER_DATA, typename NODE_TYPE>       // USER_DATA = User supplied data type, NODE_TYPE = underlying representation object (obj_list_node or ptr_obj_list_node)
        struct _linked_list : public _element_collection<USER_DATA, NODE_TYPE>
        {
            public:
            // Element access structure

                class list_element : public _element<_param_details<NODE_TYPE>, NODE_TYPE, USER_DATA>
                {
                    public:
                    // Constructors
                    list_element() : _element()                                                             { };
                    list_element(NODE_TYPE* p_obj, _param_details<NODE_TYPE>* p_in) : _element(p_obj, p_in) { };
                    list_element(const list_element& p_in) : _element(p_in)                                 { };

                    // Functions
                    void remove(const bool _do_not_destroy = false)
                    {
                        p_params.p_left     = (p_params.p_left == p_element_obj)    ? p_params.p_left.next_element()    : p_params.p_left;
                        p_params.p_right    = (p_params.p_right == p_element_obj)   ? p_params.p_right.prev_element()   : p_params.p_right;
                        p_params.p_root     = (p_params.p_root == p_element_obj)    ? p_params.p_left                   : p_params.p_root;
                        p_params._size--;
                        p_element_obj.remove(p_params.do_not_destroy || _do_not_destroy);
                        delete(p_element_obj);
                    };
                };


            protected:
            // Constructors / Destructor

                // Default constructor
                _linked_list() :
                    _element_collection<USER_DATA, NODE_TYPE>()
                { };

                // Copy constructor
                _linked_list(const _linked_list<USER_DATA, NODE_TYPE>& p_in) :
                    _element_collection<USER_DATA, NODE_TYPE>(p_in.p_params)
                { };

                // Destructor
                ~_linked_list()
                { reset(); };


            // Assignment operator

                void operator=(const _linked_list<USER_DATA, NODE_TYPE>& p_in)
                {
                    // Detach current list
                    reset();

                    // Assign linked list
                    _element_collection<USER_DATA, NODE_TYPE>::operator=(p_in);
                };


            // Comparison operator
            
                const bool operator!=(const _linked_list<USER_DATA, NODE_TYPE>& p_in)
                { return(!operator==(p_in)); };

                const bool operator==(const _linked_list<USER_DATA, NODE_TYPE>& p_in)
                { 
                    // Precheck same list
                    if(p_params == p_in.p_params)
                    { return(true); }
                    
                    // Check size
                    if(p_params._size != p_in.p_params._size)
                    { return(false); }

                    // Local init
                    int         cnt         = NULL;
                    NODE_TYPE*  local_ptr   = p_params.p_left;
                    NODE_TYPE*  remote_ptr  = p_in.p_params.p_left;
                    USER_DATA   local_obj   = local_ptr.get_obj();
                    USER_DATA   remote_obj  = remote_ptr.get_obj();

                    // Loop list
                    while( (cnt < p_params._size)
                        && (local_obj == remote_obj)
                        && (!_StopFlag) )
                    {
                        local_ptr   = local_ptr.next_element();
                        local_obj   = local_ptr.get_obj();
                        remote_ptr  = remote_ptr.next_element();
                        remote_obj  = remote_ptr.get_obj();
                        cnt++;
                    }

                    // Return
                    return((cnt == p_params._size) && (local_obj == remote_obj));
                };

            
            public:
            // Access operator

                list_element operator[](const int idx)
                { return(list_element(get_element_node(idx, p_params.p_left), p_params)); };


            // Array functions

                template <typename ARRAY_TYPE>
                const int get_array(ARRAY_TYPE& p_arr[], const bool reverse = false)                                            { return(_get_array(p_arr, NULL, WHOLE_ARRAY, reverse)); };
                template <typename ARRAY_TYPE>
                const int get_array(ARRAY_TYPE& p_arr[], const int start, const bool reverse = false)                           { return(_get_array(p_arr, start, WHOLE_ARRAY, reverse)); };
                template <typename ARRAY_TYPE>
                const int get_array(ARRAY_TYPE& p_arr[], const int start, const int end, const bool reverse = false)            { return(_get_array(p_arr, start, end, reverse)); };


            private:
            // General object functions

                // Reset linked list
                void reset()
                {
                    if(::CheckPointer(p_params) == POINTER_INVALID)
                    { return; }

                    if(p_params.instance_cnt == 1)
                    {
                        NODE_TYPE* ptr      = p_params.p_left;
                        NODE_TYPE* next_ptr = ptr;
                        for(int cnt = p_params._size - 1; (cnt >= NULL) && (ptr != NULL); cnt--)
                        {
                            if(::CheckPointer(ptr) == POINTER_DYNAMIC)
                            {
                                next_ptr    = ptr.next_element();
                                if(p_params.do_not_destroy)
                                { ptr.detach_obj(); }
                                delete(ptr);
                                ptr         = next_ptr;
                            }
                        }
                    }
                    p_params.clear();
                    p_params.instance_cnt += (p_params.instance_cnt == NULL);
                };


            // Linked list functions

                // Objects to array
                template <typename ARRAY_TYPE>
                const int _get_array(ARRAY_TYPE& p_arr[], const int start, const int end, const bool reverse)
                {
                    // Precheck
                    if(p_params.p_root == NULL)
                    {
                        ArrayFree(p_arr);
                        return(NULL);
                    }

                    // Local init
                    const int   _end        = (end != WHOLE_ARRAY) ? MathMin(end, p_params._size) : p_params._size;
                    const int   _size       = _end - start;
                    NODE_TYPE*  p_this      = (start == NULL) ? p_params.p_left : get_element_node(start, p_params.p_left);
                    int         arr_ptr     = (reverse) ? ((_end - start) - 1) : NULL;

                    // Adjust output array
                    ArrayResize(p_arr, _size);

                    // Fetch elements
                    for(int cnt = start; (cnt < _end) && (p_this != NULL) && (!_StopFlag); cnt++)
                    {
                        // Get value
                        cast_obj(p_arr, p_this, arr_ptr);

                        // Loop control
                        p_this          = p_this.next_element();
                        arr_ptr        += (reverse) ? -1 : 1;
                    }

                    // Return
                    return((_StopFlag) ? NULL : ArraySize(p_arr));
                };


            // Generic cast object function

                void cast_obj(USER_DATA& p_arr[], NODE_TYPE* p_this, const int arr_ptr)                                         { p_arr[arr_ptr] = p_this.get_obj(); }
                void cast_obj(list_element& p_arr[], NODE_TYPE* p_this, const int arr_ptr)                                      { p_arr[arr_ptr].set(p_this, p_params); }
                void cast_obj(list_element*& p_arr[], NODE_TYPE* p_this, const int arr_ptr)
                {
                    // Value assignment
                    if(CheckPointer(p_arr[arr_ptr]) == POINTER_INVALID)
                    { p_arr[arr_ptr] = new list_element(p_this, p_params); }
                    else
                    { p_arr[arr_ptr].set(p_this, p_params); }
                };


            protected:
            // Element functions

                // Prepend element
                void prepend(USER_DATA& p_in)
                {
                    // Create new object
                    NODE_TYPE* p_new = new NODE_TYPE(p_in);
                    p_params._size++;
                    p_params.p_last_added       = p_new;

                    // Add as initial element
                    if(p_params.p_root == NULL)
                    {
                        p_params.p_root         = p_new;
                        p_params.p_left         = p_new;
                        p_params.p_right        = p_new;

                        // Return
                        return;
                    }

                    // Prepend to existing chain
                    p_params.p_root.prepend(p_new);
                    p_params.p_left             = p_new;

                    // Return
                    return;
                };

                // Append element
                void append(USER_DATA& p_in)
                {
                    // Create new object
                    NODE_TYPE* p_new 			= new NODE_TYPE(p_in);
                    p_params._size++;
                    p_params.p_last_added   	= p_new;

                    // Add as initial element
                    if(p_params.p_root == NULL)
                    {
                        p_params.p_root         = p_new;
                        p_params.p_left         = p_new;
                        p_params.p_right        = p_new;

                        // Return
                        return;
                    }

                    // Append to existing chain
                    p_params.p_right.append(p_new);
                    p_params.p_right            = p_new;

                    // Return
                    return;
                };

                // Insert element
                void insert(const int idx, USER_DATA& p_in)
                {
                    // Precheck
                    if(idx >= p_params._size)
                    {
                        // Append
                        append(p_in);

                        // Return
                        return;
                    }

                    // Create new object
                    NODE_TYPE* p_new        = new NODE_TYPE(p_in);
                    p_params._size++;
                    p_params.p_last_added   = p_new;

                    // Find position
                    NODE_TYPE* p_insert_at  = get_element_node(idx, p_params.p_left);
                    p_params.p_left         = (idx == NULL) ? p_new : p_params.p_left;
                    p_insert_at.insert(p_new);
                };

                // Remove element
                void remove(const int idx, const bool _do_not_destroy = false)
                {
                    // Precheck list
                    if( (p_params.p_root == NULL)
                     || (idx >= p_params._size) )
                    { return; }

                    // Get access object
                    NODE_TYPE* p_remove = get_element_node(idx, p_params.p_left);
                    p_params._size--;

                    // Update chain pointer
                    p_params.p_root         = (p_remove != p_params.p_root)         ? p_params.p_root           : ((p_remove.prev_element() != NULL) ? p_remove.prev_element() : p_remove.next_element());
                    p_params.p_left         = (p_params.p_left == p_remove)         ? p_remove.next_element()   : p_params.p_left;
                    p_params.p_right        = (p_params.p_right == p_remove)        ? p_remove.prev_element()   : p_params.p_right;
                    p_params.p_last_added   = (p_params.p_last_added == p_remove)   ? NULL                      : p_params.p_last_added;

                    // Disconnect and remove element
                    if(p_params.do_not_destroy || _do_not_destroy)
                    { p_remove.detach_obj(); }
                    delete(p_remove);

                    // Return
                    return;
                };


            private:
            // Node by index

                NODE_TYPE* get_element_node(const int idx, NODE_TYPE* ptr_obj)
                {
                    int count = NULL;
                    while( (ptr_obj.next_element() != NULL)
                        && (count < idx)
                        && (!_StopFlag) )
                    {
                        ptr_obj = ptr_obj.next_element();
                        count++;
                    }

                    // Return
                    return(ptr_obj);
                };
        };



    ///////////////////////////////////////
    //
    //  Sorted linked list object
    //

        // Base object
        template <typename USER_DATA, typename LIST_TYPE, typename NODE_TYPE>
        struct _sorted_linked_list : public LIST_TYPE
        {
            private:
            // Local storage

                NODE_TYPE* prev_find_ptr;
                NODE_TYPE* find_ptr;


            protected:
            // Constructors / Destructor

                // Default constructor
                _sorted_linked_list() :
                    LIST_TYPE       (),
                    prev_find_ptr   (NULL),
                    find_ptr        (NULL)
                { };

                // Copy constructor
                _sorted_linked_list(const _sorted_linked_list<USER_DATA, LIST_TYPE, NODE_TYPE>& p_in) :
                    LIST_TYPE       (p_in),
                    prev_find_ptr   (p_in.prev_find_ptr),
                    find_ptr        (p_in.find_ptr)
                { };

                // Destructor
                ~_sorted_linked_list()
                { }


            // Element functions

                // Add element
                const bool add(USER_DATA& p_in, const bool unique = false)
                {
                    // Get object pointer
                    USER_DATA curr_obj;
                    NODE_TYPE* ptr          = find(p_in, curr_obj);
                    
                    if(ptr != NULL)
                    { curr_obj = ptr.get_obj(); }

                    // Check for unique element
                    if( (unique)
                     && (ptr != NULL)
                     && (curr_obj == p_in) )
                    { return(false); }

                    // Append operation
                    if( (ptr == NULL)
                     || ((ptr == p_params.p_right) && (curr_obj < p_in)) )
                    {
                        _linked_list<USER_DATA, NODE_TYPE>::append(p_in);

                        // Return
                        return(true);
                    }

                    // Prepend operation
                    if( (ptr == p_params.p_left)
                     && (curr_obj > p_in) )
                    {
                        _linked_list<USER_DATA, NODE_TYPE>::prepend(p_in);

                        // Return
                        return(true);
                    }

                    // Move to insert location
                    if(curr_obj <= p_in)
                    {
                        if(ptr.next_element() != NULL)
                        { ptr = ptr.next_element(); }
                        else
                        {
                            _linked_list<USER_DATA, NODE_TYPE>::append(p_in);
                            return(true);
                        }
                    }

                    // Check for unique element
                    if( (unique)
                     && (curr_obj == p_in) )
                    { return(false); }

                    // Insert operation
                    ptr.insert(p_in);
                    p_params.p_last_added = ptr.prev_element();

                    // Set state
                    p_params._size++;

                    // Return
                    return(true);
                };

                // Remove element
                const bool remove(USER_DATA& p_in, const bool _do_not_destroy = false)
                {
                    // Check list size
                    if(p_params._size == NULL)
                    { return(false); }

                    // Get object pointer
                    NODE_TYPE* ptr = find(p_in);

                    // Validate element removal
                    if(ptr.get_obj() != p_in)
                    { return(false); }

                    // Update first/last/chain
                    p_params.p_left         = (ptr == p_params.p_left)                                  ? ptr.next_element()    : p_params.p_left;
                    p_params.p_right        = (ptr == p_params.p_right)                                 ? ptr.prev_element()    : p_params.p_right;
                    p_params.p_root         = (ptr == p_params.p_root)                                  ? p_params.p_left       : p_params.p_root;
                    p_params.p_last_added   = (p_params.p_left == NULL) && (p_params.p_right == NULL)   ? NULL                  : p_params.p_last_added;

                    // Remove element from chain
                    if(p_params.do_not_destroy || _do_not_destroy)
                    { ptr.detach_obj(); }
                    delete(ptr);

                    // Set state
                    p_params._size--;

                    // Return
                    return(true);
                };


            private:
            // Find element

                NODE_TYPE* find(USER_DATA& p_in)
                { 
                    USER_DATA p_tmp;
                    return(find(p_in, p_tmp));
                }

                NODE_TYPE* find(USER_DATA& p_in, USER_DATA& p_found)
                {
                    // Current location
                    find_ptr        = (CheckPointer(find_ptr) == POINTER_INVALID)       ? p_params.p_left : find_ptr;
                    prev_find_ptr   = (CheckPointer(prev_find_ptr) == POINTER_INVALID)  ? find_ptr        : prev_find_ptr;

                    // Current chain status
                    if(p_params.p_root == NULL)
                    { return(NULL); }

                    // Append
                    if(p_params.p_right.get_obj() <= p_in)
                    { return(p_params.p_right); }

                    // Prepend
                    if(p_params.p_left.get_obj() >= p_in)
                    { return(p_params.p_left); }

                    // Refine find_ptr by root ptr
                    if( (p_params.p_root.get_obj() < find_ptr.get_obj())
                     && (p_in < p_params.p_root.get_obj()) )
                    { find_ptr = p_params.p_root; }
                    else if( (p_params.p_root.get_obj() > find_ptr.get_obj())
                     && (p_in > p_params.p_root.get_obj()) )
                    { find_ptr = p_params.p_root; }

                    // Select closer find_ptr
                    if( (prev_find_ptr.get_obj() < find_ptr.get_obj())
                     && (p_in < prev_find_ptr.get_obj()) )
                    { find_ptr = prev_find_ptr; }
                    else if( (prev_find_ptr.get_obj() > find_ptr.get_obj())
                     && (p_in > prev_find_ptr.get_obj()) )
                    { find_ptr = prev_find_ptr; }
                    else
                    { prev_find_ptr = find_ptr; }

                    // Loop forward
                    while( (find_ptr.get_obj() < p_in)
                        && (find_ptr != p_params.p_right)
                        && (!_StopFlag) )
                    { find_ptr = find_ptr.next_element(); }

                    // Loop backward
                    while( (find_ptr.get_obj() > p_in)
                        && (find_ptr != p_params.p_left)
                        && (!_StopFlag) )
                    { find_ptr = find_ptr.prev_element(); }

                    // Get value
                    p_found = find_ptr.get_obj();

                    // Return
                    return(find_ptr);
                };
        };



//
// END MQL data structures                                                                                                                                */
//*********************************************************************************************************************************************************/
#endif // LIB_MQLPLUS_BASE_OBJ_LIST_TEMPLATES_MQH_INCLUDED
